#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-01-10
# File  : fritzbox_smarthome_button.py (check plugin)
#
#

from time import localtime, strftime
from typing import Dict

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    Service,
    register,
    Result,
    State,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import CheckResult, DiscoveryResult
from cmk.base.plugins.agent_based.utils.fritzbox_smarthome import AvmSmartHomeDevice, AvmButton, AVM_TIME_FORMAT


def discovery_fritzbox_smarthome_button_single(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if isinstance(section, AvmSmartHomeDevice):
        if isinstance(section.buttons, list):
            for button in section.buttons:
                item = button.name.split(':')[-1].strip()  # name="Button01: Top right"
                yield Service(item=item, parameters={'discovered_id': button.id})


def discovery_fritzbox_smarthome_button_multiple(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if not isinstance(section, AvmSmartHomeDevice):
        for device_id, device in section.items():
            if isinstance(device.buttons, list):
                for button in device.buttons:
                    item = button.name.split(':')[-1].strip()  # name="Button01: Top right"
                    yield Service(item=f'{device_id} {item}', parameters={'discovered_id': button.id})


def check_fritzbox_smarthome_button_single(
        item, params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if not isinstance(section, AvmSmartHomeDevice) or section.buttons is None:
        return

    button_found = None
    for button in section.buttons:
        if button.id == params['discovered_id']:
            button_found: AvmButton | None = button
            break

    if not button_found:
        return

    if button_found.last_pressed_time_stamp is not None:
        yield Result(
            state=State.OK,
            summary=f'Last pressed: {strftime(AVM_TIME_FORMAT, localtime(button_found.last_pressed_time_stamp))}'
        )
    else:
        yield Result(state=State.OK, summary='Button never pressed')

    yield Result(state=State.OK, notice=f'ID: {button_found.id}')
    yield Result(state=State.OK, notice=f'Identifier: {button_found.identifier}')


def check_fritzbox_smarthome_button_multiple(
        item, params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if isinstance(section, Dict):
        try:
            yield from check_fritzbox_smarthome_button_single(item, params, section[item])
        except KeyError:
            return


register.check_plugin(
    name='fritzbox_smarthome_button_single',
    service_name='Button %s',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_button_single,
    check_function=check_fritzbox_smarthome_button_single,
    # check_ruleset_name='fritzbox_smarthome_button_single',
    check_default_parameters={}
)


register.check_plugin(
    name='fritzbox_smarthome_button_multiple',
    service_name='Smarthome Button %s',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_button_multiple,
    check_function=check_fritzbox_smarthome_button_multiple,
    # check_ruleset_name='fritzbox_smarthome_button_multiple',
    check_default_parameters={}
)
