#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-02-23
# File  : fritzbox_smarthome_sensor.py (check plugin)
#
#

from time import localtime, strftime
from typing import Dict

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    Service,
    register,
    Result,
    State,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import CheckResult, DiscoveryResult
from cmk.base.plugins.agent_based.utils.fritzbox_smarthome import AvmSmartHomeDevice, AVM_TIME_FORMAT


def discovery_fritzbox_smarthome_sensor_single(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if isinstance(section, AvmSmartHomeDevice):
        if section.alert is not None:
            yield Service()


def discovery_fritzbox_smarthome_sensor_multiple(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if not isinstance(section, AvmSmartHomeDevice):
        for device_id, device in section.items():
            if device.alert is not None:
                yield Service(item=str(device_id))


def check_fritzbox_smarthome_sensor_single(
        params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if not isinstance(section, AvmSmartHomeDevice) or section.alert is None:
        return

    def _get_status(status: int):
        _state = {
            0: ('closed', params.get('state', {}).get('closed', 0)),
            1: ('open', params.get('state', {}).get('open', 0)),
        }
        return _state.get(status, (f'unknown ({status})', 3))

    state_readable, state = _get_status(section.alert.state)
    yield Result(state=State(state), summary=f'State: {state_readable}')
    if section.alert.last_changed_time_stamp is not None:
        yield Result(
            state=State.OK,
            summary=f'Last changed: {strftime(AVM_TIME_FORMAT, localtime(section.alert.last_changed_time_stamp))}'
        )


def check_fritzbox_smarthome_sensor_multiple(
        item, params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if isinstance(section, Dict):
        try:
            yield from check_fritzbox_smarthome_sensor_single(params, section[item])
        except KeyError:
            return


register.check_plugin(
    name='fritzbox_smarthome_sensor_single',
    service_name='Sensor',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_sensor_single,
    check_function=check_fritzbox_smarthome_sensor_single,
    check_ruleset_name='fritzbox_smarthome_sensor_single',
    check_default_parameters={}
)


register.check_plugin(
    name='fritzbox_smarthome_sensor_multiple',
    service_name='Smarthome Sensor %s',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_sensor_multiple,
    check_function=check_fritzbox_smarthome_sensor_multiple,
    check_ruleset_name='fritzbox_smarthome_sensor_multiple',
    check_default_parameters={}
)
