#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-12-29
# File  : fritzbox_smarthome.py (check plugin utils)
#
# Based on the work of Maximilian Clemens, see https://github.com/MaximilianClemens/checkmk_fritzbox
#
#

from dataclasses import dataclass
from typing import Any, Dict, Final, List
from os import environ
from pathlib import Path
from json import loads, dumps

from cmk.checkers import plugin_contexts


@dataclass(frozen=True)
class AvmButton:
    identifier: str
    id: str
    name: str
    last_pressed_time_stamp: int


@dataclass(frozen=True)
class AvmHumidity:
    rel_humidity: int


@dataclass(frozen=True)
class AvmTemperature:
    celsius: float | None
    offset: float | None


@dataclass(frozen=True)
class AvmPowerMeter:
    energy: float | None
    power: float | None
    voltage: float | None


@dataclass(frozen=True)
class AvmSimpleOnOff:
    state: int | None


@dataclass(frozen=True)
class AvmNextChange:
    end_period: int | None
    temp_change_to: float | None


@dataclass(frozen=True)
class AvmThermostat:
    error_code: int | None
    temp_comfort: float | None
    temp_current: float | None
    temp_economic: float | None
    temp_target: float | None
    next_change: AvmNextChange | None = None
    adaptive_heating_active: int | None = None
    adaptive_heating_running: int | None = None
    boost_active: int | None = None
    boost_active_end_time: int | None = None
    holiday_active: int | None = None
    summer_active: int | None = None
    window_open_activ: int | None = None
    window_open_active_end_time: int | None = None


@dataclass(frozen=True)
class AvmSwitch:
    mode: str
    state: int | None


@dataclass(frozen=True)
class AvmUnit:
    device_id: str
    unit_type: int
    interfaces: int


@dataclass(frozen=True)
class AvmAlert:
    state: int
    last_changed_time_stamp: int


@dataclass(frozen=True)
class AvmSmartHomeDevice:
    fbm: int | None
    functions: List[str]
    fw_version: str
    id: str
    identifier: str
    manufacturer: str
    name: str
    present: int
    product_name: str
    battery: int | None = None
    battery_low: int | None = None
    device_lock: int | None = None
    lock: int | None = None
    power_meter: AvmPowerMeter | None = None
    simple_on_off: AvmSimpleOnOff | None = None
    switch: AvmSwitch | None = None
    temperature: AvmTemperature | None = None
    thermostat: AvmThermostat | None = None
    tx_busy: int | None = None
    buttons: list[AvmButton] | None = None
    humidity: AvmHumidity | None = None
    unit: AvmUnit | None = None
    alert: AvmAlert | None = None


_AVM_ADAPTIVE_HEATING_ACTIVE = 'adaptiveHeatingActive'
_AVM_ADAPTIVE_HEATING_RUNNING = 'adaptiveHeatingRunning'
_AVM_ALERT = 'alert'
_AVM_BATTERY = 'battery'
_AVM_BATTERY_LOW = 'batterylow'
_AVM_BOOST_ACTIVE = 'boostactive'
_AVM_BOOST_ACTIVE_END_TIME = 'boostactiveendtime'
_AVM_CELSIUS = 'celsius'
_AVM_DEVICE_LOCK = 'devicelock'
_AVM_END_PERIOD = 'endperiod'
_AVM_ENERGY = 'energy'
_AVM_ERROR_CODE = 'errorcode'
_AVM_ETSI_DEVICE_ID = 'etsideviceid'
_AVM_ETSI_UNIT_INFO = "etsiunitinfo"
_AVM_FUNCTION_BIT_MASK = 'functionbitmask'
_AVM_FW_REVISION = 'fwversion'
_AVM_HOLIDAY_ACTIVE = 'holidayactive'
_AVM_HUMIDITY = 'humidity'
_AVM_ID = 'id'
_AVM_IDENTIFIER = 'identifier'
_AVM_INTERFACES = 'interfaces'
_AVM_LAST_ALERT_CHG_TIME_STAMP = 'lastalertchgtimestamp'
_AVM_LAST_PRESSED_TIME_STAMP = 'lastpressedtimestamp'
_AVM_LOCK = 'lock'
_AVM_MANUFACTURER = 'manufacturer'
_AVM_MODE = 'mode'
_AVM_NAME='name'
_AVM_NEXT_CHANGE = 'nextchange'
_AVM_OFFSET = 'offset'
_AVM_POWER = 'power'
_AVM_POWER_METER = 'powermeter'
_AVM_PRESENT = 'present'
_AVM_PRODUCT_NAME = 'productname'
_AVM_REL_HUMIDITY = 'rel_humidity'
_AVM_SIMPLE_ON_OFF = 'simpleonoff'
_AVM_STATE = 'state'
_AVM_SUMMER_ACTIVE = 'summeractive'
_AVM_SWITCH = 'switch'
_AVM_TEMPERATURE = 'temperature'
_AVM_TEMP_CHANGE = 'tchange'
_AVM_TEMP_COMFORT = 'komfort'
_AVM_TEMP_CURRENT = 'tist'
_AVM_TEMP_ECONOMIC = 'absenk'
_AVM_TEMP_TARGET = 'tsoll'
_AVM_THERMOSTAT = 'hkr'
_AVM_TX_BUSY = 'txbusy'
_AVM_UNIT_TYPE = 'unittype'
_AVM_VOLTAGE = 'voltage'
_AVM_WINDOW_OPEN_ACTIV = 'windowopenactiv'
_AVM_WINDOW_OPEN_ACTIVE_END_TIME = 'windowopenactiveendtime'

_OMD_ROOT = environ["OMD_ROOT"]

AVM_TIME_FORMAT = '%Y-%m-%dT%H:%M:%S'

HAN_FUN_UNIT_TYPE: Final = {
    273: 'Simple Button',
    256: 'Simple on/off switchable',
    257: 'Simple on//ff switch',
    262: 'AC outlet',
    263: 'AC outlet, simple power metering',
    264: 'Simple light',
    265: 'Dimmable light ',
    266: 'Dimmer switch ',
    277: 'Colour blub',
    278: 'Dimmable color blub',
    281: 'Blind',
    282: 'Lamellar',
    512: 'Simple detector',
    513: 'Door open/close detector',
    514: 'Window open/close detector',
    515: 'Motion detector',
    518: 'Flood detector',
    519: 'Glas break detector',
    520: 'Vibration detector',
    640: 'Siren',

}

HAN_FUN_INTERFACE: Final = {
    277: 'Keep alive',
    256: 'Alert',
    512: 'On/off',
    513: 'Level control',
    514: 'color control',
    516: 'Open/cloe',
    517: 'Open/close config',
    772: 'Simple button',
    1024: 'SUOTA Update',
}


class FritzBoxValueStore:
    """
    Persistent kind of value_store. Located under ~/var/fritzbox_smarthome/{hostname}_{service_description}
    """
    def __init__(self):
        self._host_name = plugin_contexts.host_name()
        self._service_description = plugin_contexts.service_description().replace(' ', '_')
        self._file = Path(f'{_OMD_ROOT}/var/firtzbox_smarthome/{self._host_name}_{self._service_description}')
        self._counters = {}
        if self._file.exists():
            self._counters = loads(self._file.read_text())
        else:
            self._counters = {}

    def get(self, key: str, default_value: int | str | float | bool | None = None):
        """
        Get a value from the value store
        Args:
            key: the key in the value store
            default_value: the default value to return if key not found

        Returns:
            The value for key from the value store. If key not exists the default_value or None
        """
        return self._counters.get(key, default_value)

    def set(self, key: str, value: str | int | float | None) -> None:
        """
        Set a value in the value stare. Use "None" as value to remove a key.
        Args:
            key: the key in the value store
            value: the value to set
        """
        if value is None:
            try:
                self._counters.pop(key)
            except KeyError:
                pass
        else:
            self._counters[key] = value

    def save(self) -> None:
        """
        Save the value_store as json to file
        """
        self._file.parent.mkdir(exist_ok=True, parents=True)
        self._file.write_text(dumps(self._counters))


def _get_battery(device: Dict[str, Any]) -> int | None:
    try:
        return int(device[_AVM_THERMOSTAT][_AVM_BATTERY])
    except (KeyError, ValueError):
        pass

    try:
        return int(device[_AVM_BATTERY])
    except (KeyError, ValueError):
        pass


def _get_battery_low(device: Dict[str, Any]) -> int | None:
    try:
        return int(device[_AVM_THERMOSTAT][_AVM_BATTERY_LOW])
    except (KeyError, ValueError):
        pass

    try:
        return int(device[_AVM_BATTERY_LOW])
    except (KeyError, ValueError):
        pass


def _get_lock(device: Dict[str, Any]) -> int | None:
    try:
        return int(device[_AVM_THERMOSTAT][_AVM_LOCK])
    except (KeyError, ValueError):
        pass

    try:
        return int(device[_AVM_SWITCH][_AVM_LOCK])
    except (KeyError, ValueError):
        pass


def _get_device_lock(device: Dict[str, Any]) -> int | None:
    try:
        return int(device[_AVM_THERMOSTAT][_AVM_DEVICE_LOCK])
    except (KeyError, ValueError):
        pass

    try:
        return int(device[_AVM_SWITCH][_AVM_DEVICE_LOCK])
    except (KeyError, ValueError):
        pass


def _get_buttons(device: Dict[str, any]) -> List[AvmButton] | None:
    return [
        AvmButton(
            identifier=device[key][_AVM_IDENTIFIER],
            id=device[key][_AVM_ID],
            name=device[key][_AVM_NAME],
            last_pressed_time_stamp=_get_int(device[key][_AVM_LAST_PRESSED_TIME_STAMP])
        ) for key in device.keys() if key.startswith('button')
    ]


def _get_int(value: str | None) -> int | None:
    try:
        return int(value)
    except (ValueError, TypeError):
        return


def _get_float(value: str | None, scale: float = 1.0) -> float | None:
    try:
        return float(value) / scale
    except (ValueError, TypeError):
        return


def parse_avm_smarthome_device(raw_device: Dict[str, Any]) -> AvmSmartHomeDevice:
    return AvmSmartHomeDevice(
        battery=_get_battery(raw_device),
        battery_low=_get_battery_low(raw_device),
        device_lock=_get_device_lock(raw_device),
        fbm=_get_int(raw_device.get(_AVM_FUNCTION_BIT_MASK)),
        functions=get_avm_device_functions_from_fbm(_get_int(raw_device.get(_AVM_FUNCTION_BIT_MASK))),
        fw_version=str(raw_device[_AVM_FW_REVISION]),
        id=str(raw_device[_AVM_ID]),
        identifier=str(raw_device[_AVM_IDENTIFIER]),
        lock=_get_lock(raw_device),
        manufacturer=str(raw_device[_AVM_MANUFACTURER]),
        name=str(raw_device[_AVM_NAME]),
        present=_get_int(raw_device.get(_AVM_PRESENT)),
        product_name=str(raw_device[_AVM_PRODUCT_NAME]),
        tx_busy=_get_int(raw_device.get(_AVM_TX_BUSY)),
        temperature=AvmTemperature(
                celsius=_get_float(value=raw_device[_AVM_TEMPERATURE].get(_AVM_CELSIUS), scale=10.0),
                offset=_get_float(value=raw_device[_AVM_TEMPERATURE].get(_AVM_OFFSET), scale=10.0),
        ) if raw_device.get(_AVM_TEMPERATURE) else None,
        thermostat=AvmThermostat(
            temp_current=_get_float(value=raw_device[_AVM_THERMOSTAT].get(_AVM_TEMP_CURRENT), scale=2.0),
            temp_target=_get_float(value=raw_device[_AVM_THERMOSTAT].get(_AVM_TEMP_TARGET), scale=2.0),
            temp_economic=_get_float(value=raw_device[_AVM_THERMOSTAT].get(_AVM_TEMP_ECONOMIC), scale=2.0),
            temp_comfort=_get_float(value=raw_device[_AVM_THERMOSTAT].get(_AVM_TEMP_COMFORT), scale=2.0),
            error_code=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_ERROR_CODE)),
            next_change=AvmNextChange(
                end_period=_get_int(raw_device[_AVM_THERMOSTAT][_AVM_NEXT_CHANGE].get(_AVM_END_PERIOD)),
                temp_change_to=_get_float(
                    value=raw_device[_AVM_THERMOSTAT][_AVM_NEXT_CHANGE].get(_AVM_TEMP_CHANGE), scale=2.0
                ),
            ) if raw_device[_AVM_THERMOSTAT].get(_AVM_NEXT_CHANGE) else None,
            adaptive_heating_active=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_ADAPTIVE_HEATING_ACTIVE)),
            adaptive_heating_running=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_ADAPTIVE_HEATING_RUNNING)),
            boost_active=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_BOOST_ACTIVE)),
            boost_active_end_time=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_BOOST_ACTIVE_END_TIME)),
            holiday_active=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_HOLIDAY_ACTIVE)),
            summer_active=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_SUMMER_ACTIVE)),
            window_open_activ=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_WINDOW_OPEN_ACTIV)),
            window_open_active_end_time=_get_int(value=raw_device[_AVM_THERMOSTAT].get(_AVM_WINDOW_OPEN_ACTIVE_END_TIME)),
        ) if raw_device.get(_AVM_THERMOSTAT) else None,
        switch=AvmSwitch(
            state=_get_int(raw_device[_AVM_SWITCH].get(_AVM_STATE)),
            mode=str(raw_device[_AVM_SWITCH][_AVM_MODE]),
        ) if raw_device.get(_AVM_SWITCH) else None,
        power_meter=AvmPowerMeter(
            voltage=_get_float(raw_device[_AVM_POWER_METER].get(_AVM_VOLTAGE),1000),
            power=_get_float(raw_device[_AVM_POWER_METER].get(_AVM_POWER), 1000),
            energy=_get_float(raw_device[_AVM_POWER_METER].get(_AVM_ENERGY)),  # / 1000,
        ) if raw_device.get(_AVM_POWER_METER) else None,
        simple_on_off=AvmSimpleOnOff(
            state=_get_int(raw_device[_AVM_SIMPLE_ON_OFF].get(_AVM_STATE)),
        ) if raw_device.get(_AVM_SIMPLE_ON_OFF) else None,
        humidity=AvmHumidity(
            rel_humidity=_get_int(raw_device[_AVM_HUMIDITY].get(_AVM_REL_HUMIDITY))
        ) if raw_device.get(_AVM_HUMIDITY) else None,
        buttons=_get_buttons(raw_device),
        unit=AvmUnit(
            device_id=raw_device[_AVM_ETSI_UNIT_INFO][_AVM_ETSI_DEVICE_ID],
            unit_type=_get_int(raw_device[_AVM_ETSI_UNIT_INFO][_AVM_UNIT_TYPE]),
            interfaces=_get_int(raw_device[_AVM_ETSI_UNIT_INFO][_AVM_INTERFACES]),
        ) if raw_device.get(_AVM_ETSI_UNIT_INFO) else None,
        alert=AvmAlert(
            state=_get_int(raw_device[_AVM_ALERT].get(_AVM_STATE)),
            last_changed_time_stamp=_get_int(raw_device[_AVM_ALERT].get(_AVM_LAST_ALERT_CHG_TIME_STAMP))
        ) if raw_device.get(_AVM_ALERT) else None
    )


def get_avm_device_functions_from_fbm(fbm: int | None) -> List[str]:
    functions = []
    if fbm is None:
        return functions

    if fbm >> 0 & 1:
        functions.append('HAN-FUN Device')
    if fbm >> 2 & 1:
        functions.append('Light')
    if fbm >> 3 & 1:
        functions.append('unknown (Bit 3)')
    if fbm >> 4 & 1:
        functions.append('Alarm Sensor')
    if fbm >> 5 & 1:
        functions.append('AVM Button')
    if fbm >> 6 & 1:
        functions.append('AVM Thermostat')
    if fbm >> 7 & 1:
        functions.append('AVM Powermeter')
    if fbm >> 8 & 1:
        functions.append('Temperature Sensor')
    if fbm >> 9 & 1:
        functions.append('AVM Switching socket')
    if fbm >> 10 & 1:
        functions.append('AVM DECT Repeater')
    if fbm >> 11 & 1:
        functions.append('AVM Microphone')
    if fbm >> 12 & 1:
        functions.append('unknown (Bit 12)')
    if fbm >> 13 & 1:
        functions.append('HAN-FUN Unit')
    if fbm >> 14 & 1:
        functions.append('unknown (Bit 14)')
    if fbm >> 15 & 1:
        functions.append('on/off switchable device')
    if fbm >> 16 & 1:
        functions.append('Device with adjustable dimming, height and level')
    if fbm >> 17 & 1:
        functions.append('Light')
    if fbm >> 18 & 1:
        functions.append('Roller shutter')
    if fbm >> 19 & 1:
        functions.append('unknown (Bit 19)')
    if fbm >> 20 & 1:
        functions.append('Humidity sensor')

    functions.sort()

    return functions
